@@ -1,24 +1,7 @@
-# The MIT License (MIT)
+# SPDX-FileCopyrightText: 2017 Tony DiCola for Adafruit Industries
 #
-# Copyright (c) 2017 Tony DiCola for Adafruit Industries
-#
-# Permission is hereby granted, free of charge, to any person obtaining a copy
-# of this software and associated documentation files (the "Software"), to deal
-# in the Software without restriction, including without limitation the rights
-# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-# copies of the Software, and to permit persons to whom the Software is
-# furnished to do so, subject to the following conditions:
-#
-# The above copyright notice and this permission notice shall be included in
-# all copies or substantial portions of the Software.
-#
-# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-# THE SOFTWARE.
+# SPDX-License-Identifier: MIT
+
 """
 `adafruit_rfm9x`
 ====================================================
@@ -29,38 +12,27 @@ http: www.airspayce.com/mikem/arduino/RadioHead/
 
 * Author(s): Tony DiCola, Jerry Needell
 """
+import random
 import time
-import digitalio
+import adafruit_bus_device.spi_device as spidev
 from micropython import const
 
+HAS_SUPERVISOR = False
+
 try:
-    from warnings import warn
+    import supervisor
+
+    if hasattr(supervisor, "ticks_ms"):
+        HAS_SUPERVISOR = True
 except ImportError:
-
-    def warn(msg, **kwargs):
-        "Issue a warning to stdout."
-        print(
-            "%s: %s"
-            % ("Warning" if kwargs.get("cat") is None else kwargs["cat"].__name__, msg)
-        )
-
-
-import adafruit_bus_device.spi_device as spidev
-
-
+    pass
 __version__ = "0.0.0-auto.0"
 __repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_RFM9x.git"
 
-
-# pylint: disable=bad-whitespace
 # Internal constants:
 # Register names (FSK Mode even though we use LoRa instead, from table 85)
 _RH_RF95_REG_00_FIFO = const(0x00)
 _RH_RF95_REG_01_OP_MODE = const(0x01)
-_RH_RF95_REG_02_RESERVED = const(0x02)
-_RH_RF95_REG_03_RESERVED = const(0x03)
-_RH_RF95_REG_04_RESERVED = const(0x04)
-_RH_RF95_REG_05_RESERVED = const(0x05)
 _RH_RF95_REG_06_FRF_MSB = const(0x06)
 _RH_RF95_REG_07_FRF_MID = const(0x07)
 _RH_RF95_REG_08_FRF_LSB = const(0x08)
@@ -107,118 +79,9 @@ _RH_RF95_REG_62_AGC_THRESH1 = const(0x62)
 _RH_RF95_REG_63_AGC_THRESH2 = const(0x63)
 _RH_RF95_REG_64_AGC_THRESH3 = const(0x64)
 
-# RH_RF95_REG_01_OP_MODE                             0x01
-_RH_RF95_LONG_RANGE_MODE = const(0x80)
-_RH_RF95_ACCESS_SHARED_REG = const(0x40)
-_RH_RF95_MODE = const(0x07)
-_RH_RF95_MODE_SLEEP = const(0x00)
-_RH_RF95_MODE_STDBY = const(0x01)
-_RH_RF95_MODE_FSTX = const(0x02)
-_RH_RF95_MODE_TX = const(0x03)
-_RH_RF95_MODE_FSRX = const(0x04)
-_RH_RF95_MODE_RXCONTINUOUS = const(0x05)
-_RH_RF95_MODE_RXSINGLE = const(0x06)
-_RH_RF95_MODE_CAD = const(0x07)
-
-# RH_RF95_REG_09_PA_CONFIG                           0x09
-_RH_RF95_PA_SELECT = const(0x80)
-_RH_RF95_MAX_POWER = const(0x70)
-_RH_RF95_OUTPUT_POWER = const(0x0F)
-
-# RH_RF95_REG_0A_PA_RAMP                             0x0a
-_RH_RF95_LOW_PN_TX_PLL_OFF = const(0x10)
-_RH_RF95_PA_RAMP = const(0x0F)
-_RH_RF95_PA_RAMP_3_4MS = const(0x00)
-_RH_RF95_PA_RAMP_2MS = const(0x01)
-_RH_RF95_PA_RAMP_1MS = const(0x02)
-_RH_RF95_PA_RAMP_500US = const(0x03)
-_RH_RF95_PA_RAMP_250US = const(0x04)
-_RH_RF95_PA_RAMP_125US = const(0x05)
-_RH_RF95_PA_RAMP_100US = const(0x06)
-_RH_RF95_PA_RAMP_62US = const(0x07)
-_RH_RF95_PA_RAMP_50US = const(0x08)
-_RH_RF95_PA_RAMP_40US = const(0x09)
-_RH_RF95_PA_RAMP_31US = const(0x0A)
-_RH_RF95_PA_RAMP_25US = const(0x0B)
-_RH_RF95_PA_RAMP_20US = const(0x0C)
-_RH_RF95_PA_RAMP_15US = const(0x0D)
-_RH_RF95_PA_RAMP_12US = const(0x0E)
-_RH_RF95_PA_RAMP_10US = const(0x0F)
-
-# RH_RF95_REG_0B_OCP                                 0x0b
-_RH_RF95_OCP_ON = const(0x20)
-_RH_RF95_OCP_TRIM = const(0x1F)
-
-# RH_RF95_REG_0C_LNA                                 0x0c
-_RH_RF95_LNA_GAIN = const(0xE0)
-_RH_RF95_LNA_BOOST = const(0x03)
-_RH_RF95_LNA_BOOST_DEFAULT = const(0x00)
-_RH_RF95_LNA_BOOST_150PC = const(0x11)
-
-# RH_RF95_REG_11_IRQ_FLAGS_MASK                      0x11
-_RH_RF95_RX_TIMEOUT_MASK = const(0x80)
-_RH_RF95_RX_DONE_MASK = const(0x40)
-_RH_RF95_PAYLOAD_CRC_ERROR_MASK = const(0x20)
-_RH_RF95_VALID_HEADER_MASK = const(0x10)
-_RH_RF95_TX_DONE_MASK = const(0x08)
-_RH_RF95_CAD_DONE_MASK = const(0x04)
-_RH_RF95_FHSS_CHANGE_CHANNEL_MASK = const(0x02)
-_RH_RF95_CAD_DETECTED_MASK = const(0x01)
-
-# RH_RF95_REG_12_IRQ_FLAGS                           0x12
-_RH_RF95_RX_TIMEOUT = const(0x80)
-_RH_RF95_RX_DONE = const(0x40)
-_RH_RF95_PAYLOAD_CRC_ERROR = const(0x20)
-_RH_RF95_VALID_HEADER = const(0x10)
-_RH_RF95_TX_DONE = const(0x08)
-_RH_RF95_CAD_DONE = const(0x04)
-_RH_RF95_FHSS_CHANGE_CHANNEL = const(0x02)
-_RH_RF95_CAD_DETECTED = const(0x01)
-
-# RH_RF95_REG_18_MODEM_STAT                          0x18
-_RH_RF95_RX_CODING_RATE = const(0xE0)
-_RH_RF95_MODEM_STATUS_CLEAR = const(0x10)
-_RH_RF95_MODEM_STATUS_HEADER_INFO_VALID = const(0x08)
-_RH_RF95_MODEM_STATUS_RX_ONGOING = const(0x04)
-_RH_RF95_MODEM_STATUS_SIGNAL_SYNCHRONIZED = const(0x02)
-_RH_RF95_MODEM_STATUS_SIGNAL_DETECTED = const(0x01)
-
-# RH_RF95_REG_1C_HOP_CHANNEL                         0x1c
-_RH_RF95_PLL_TIMEOUT = const(0x80)
-_RH_RF95_RX_PAYLOAD_CRC_IS_ON = const(0x40)
-_RH_RF95_FHSS_PRESENT_CHANNEL = const(0x3F)
-
-# RH_RF95_REG_1D_MODEM_CONFIG1                       0x1d
-_RH_RF95_BW = const(0xC0)
-_RH_RF95_BW_125KHZ = const(0x00)
-_RH_RF95_BW_250KHZ = const(0x40)
-_RH_RF95_BW_500KHZ = const(0x80)
-_RH_RF95_BW_RESERVED = const(0xC0)
-_RH_RF95_CODING_RATE = const(0x38)
-_RH_RF95_CODING_RATE_4_5 = const(0x00)
-_RH_RF95_CODING_RATE_4_6 = const(0x08)
-_RH_RF95_CODING_RATE_4_7 = const(0x10)
-_RH_RF95_CODING_RATE_4_8 = const(0x18)
-_RH_RF95_IMPLICIT_HEADER_MODE_ON = const(0x04)
-_RH_RF95_RX_PAYLOAD_CRC_ON = const(0x02)
-_RH_RF95_LOW_DATA_RATE_OPTIMIZE = const(0x01)
-
-# RH_RF95_REG_1E_MODEM_CONFIG2                       0x1e
 _RH_RF95_DETECTION_OPTIMIZE = const(0x31)
 _RH_RF95_DETECTION_THRESHOLD = const(0x37)
-_RH_RF95_SPREADING_FACTOR = const(0xF0)
-_RH_RF95_SPREADING_FACTOR_64CPS = const(0x60)
-_RH_RF95_SPREADING_FACTOR_128CPS = const(0x70)
-_RH_RF95_SPREADING_FACTOR_256CPS = const(0x80)
-_RH_RF95_SPREADING_FACTOR_512CPS = const(0x90)
-_RH_RF95_SPREADING_FACTOR_1024CPS = const(0xA0)
-_RH_RF95_SPREADING_FACTOR_2048CPS = const(0xB0)
-_RH_RF95_SPREADING_FACTOR_4096CPS = const(0xC0)
-_RH_RF95_TX_CONTINUOUS_MOE = const(0x08)
-_RH_RF95_AGC_AUTO_ON = const(0x04)
-_RH_RF95_SYM_TIMEOUT_MSB = const(0x03)
 
-# RH_RF95_REG_4D_PA_DAC                              0x4d
 _RH_RF95_PA_DAC_DISABLE = const(0x04)
 _RH_RF95_PA_DAC_ENABLE = const(0x07)
 
@@ -231,6 +94,12 @@ _RH_RF95_FSTEP = _RH_RF95_FXOSC / 524288
 # RadioHead specific compatibility constants.
 _RH_BROADCAST_ADDRESS = const(0xFF)
 
+# The acknowledgement bit in the FLAGS
+# The top 4 bits of the flags are reserved for RadioHead. The lower 4 bits are reserved
+# for application layer use.
+_RH_FLAGS_ACK = const(0x80)
+_RH_FLAGS_RETRY = const(0x40)
+
 # User facing constants:
 SLEEP_MODE = 0b000
 STANDBY_MODE = 0b001
@@ -238,14 +107,26 @@ FS_TX_MODE = 0b010
 TX_MODE = 0b011
 FS_RX_MODE = 0b100
 RX_MODE = 0b101
-# pylint: enable=bad-whitespace
-
+# supervisor.ticks_ms() contants
+_TICKS_PERIOD = const(1 << 29)
+_TICKS_MAX = const(_TICKS_PERIOD - 1)
+_TICKS_HALFPERIOD = const(_TICKS_PERIOD // 2)
 
 # Disable the too many instance members warning.  Pylint has no knowledge
 # of the context and is merely guessing at the proper amount of members.  This
 # is a complex chip which requires exposing many attributes and state.  Disable
 # the warning to work around the error.
 # pylint: disable=too-many-instance-attributes
+# pylint: disable=too-many-statements
+
+
+def ticks_diff(ticks1, ticks2):
+    """Compute the signed difference between two ticks values
+    assuming that they are within 2**28 ticks
+    """
+    diff = (ticks1 - ticks2) & _TICKS_MAX
+    diff = ((diff + _TICKS_HALFPERIOD) & _TICKS_MAX) - _TICKS_HALFPERIOD
+    return diff
 
 
 class RFM9x:
@@ -265,7 +146,8 @@ class RFM9x:
     is True for high power.
     - baudrate: Baud rate of the SPI connection, default is 10mhz but you might
     choose to lower to 1mhz if using long wires or a breadboard.
-
+    - agc: Boolean to Enable/Disable Automatic Gain Control - Default=False (AGC off)
+    - crc: Boolean to Enable/Disable Cyclic Redundancy Check - Default=True (CRC Enabled)
     Remember this library makes a best effort at receiving packets with pure
     Python code.  Trying to receive packets too quickly will result in lost data
     so limit yourself to simple scenarios of sending and receiving single
@@ -273,15 +155,16 @@ class RFM9x:
 
     Also note this library tries to be compatible with raw RadioHead Arduino
     library communication. This means the library sets up the radio modulation
-    to match RadioHead's defaults. Features like addressing and guaranteed
-    delivery need to be implemented at an application level.
+    to match RadioHead's defaults and assumes that each packet contains a
+    4 byte header compatible with RadioHead's implementation.
+    Advanced RadioHead features like address/node specific packets
+    or "reliable datagram" delivery are supported however due to the
+    limitations noted, "reliable datagram" is still subject to missed packets but with it,
+    sender is notified if a packet has potentially been missed.
     """
 
-    # Global buffer to hold data sent and received with the chip.  This must be
-    # at least as large as the FIFO on the chip (256 bytes)!  Keep this on the
-    # class level to ensure only one copy ever exists (with the trade-off that
-    # this is NOT re-entrant or thread safe code by design).
-    _BUFFER = bytearray(10)
+    # Global buffer for SPI commands
+    _BUFFER = bytearray(4)
 
     class _RegisterBits:
         # Class to simplify access to the many configuration bits avaialable
@@ -344,11 +227,17 @@ class RFM9x:
 
     dio0_mapping = _RegisterBits(_RH_RF95_REG_40_DIO_MAPPING1, offset=6, bits=2)
 
-    tx_done = _RegisterBits(_RH_RF95_REG_12_IRQ_FLAGS, offset=3, bits=1)
+    auto_agc = _RegisterBits(_RH_RF95_REG_26_MODEM_CONFIG3, offset=2, bits=1)
 
-    rx_done = _RegisterBits(_RH_RF95_REG_12_IRQ_FLAGS, offset=6, bits=1)
+    low_datarate_optimize = _RegisterBits(
+        _RH_RF95_REG_26_MODEM_CONFIG3, offset=3, bits=1
+    )
 
-    crc_error = _RegisterBits(_RH_RF95_REG_12_IRQ_FLAGS, offset=5, bits=1)
+    lna_boost_hf = _RegisterBits(_RH_RF95_REG_0C_LNA, offset=0, bits=2)
+
+    auto_ifon = _RegisterBits(_RH_RF95_DETECTION_OPTIMIZE, offset=7, bits=1)
+
+    detection_optimize = _RegisterBits(_RH_RF95_DETECTION_OPTIMIZE, offset=0, bits=3)
 
     bw_bins = (7800, 10400, 15600, 20800, 31250, 41700, 62500, 125000, 250000)
 
@@ -361,18 +250,19 @@ class RFM9x:
         *,
         preamble_length=8,
         high_power=True,
-        baudrate=5000000
+        baudrate=5000000,
+        agc=False,
+        crc=True
     ):
         self.high_power = high_power
         # Device support SPI mode 0 (polarity & phase = 0) up to a max of 10mhz.
         # Set Default Baudrate to 5MHz to avoid problems
         self._device = spidev.SPIDevice(spi, cs, baudrate=baudrate, polarity=0, phase=0)
-        # Setup reset as a digital input (default state for reset line according
-        # to the datasheet).  This line is pulled low as an output quickly to
-        # trigger a reset.  Note that reset MUST be done like this and set as
-        # a high impedence input or else the chip cannot change modes (trust me!).
+        # Setup reset as a digital output - initially High
+        # This line is pulled low as an output quickly to trigger a reset.
         self._reset = reset
-        self._reset.switch_to_input(pull=digitalio.Pull.UP)
+        # initialize Reset High
+        self._reset.switch_to_output(value=True)
         self.reset()
         # No device type check!  Catch an error from the very first request and
         # throw a nicer message to indicate possible wiring problems.
@@ -397,20 +287,79 @@ class RFM9x:
         self._write_u8(_RH_RF95_REG_0F_FIFO_RX_BASE_ADDR, 0x00)
         # Set mode idle
         self.idle()
+        # Set frequency
+        self.frequency_mhz = frequency
+        # Set preamble length (default 8 bytes to match radiohead).
+        self.preamble_length = preamble_length
         # Defaults set modem config to RadioHead compatible Bw125Cr45Sf128 mode.
         self.signal_bandwidth = 125000
         self.coding_rate = 5
         self.spreading_factor = 7
-        # Default to disable CRC checking on incoming packets.
-        self.enable_crc = False
-        # Note no sync word is set for LoRa mode either!
-        self._write_u8(_RH_RF95_REG_26_MODEM_CONFIG3, 0x00)  # Preamble lsb?
-        # Set preamble length (default 8 bytes to match radiohead).
-        self.preamble_length = preamble_length
-        # Set frequency
-        self.frequency_mhz = frequency
-        # Set TX power to low defaut, 13 dB.
+        # Default to enable CRC checking on incoming packets.
+        self.enable_crc = crc
+        """CRC Enable state"""
+        # set AGC - Default = False
+        self.auto_agc = agc
+        """Automatic Gain Control state"""
+        # Set transmit power to 13 dBm, a safe value any module supports.
         self.tx_power = 13
+        # initialize last RSSI reading
+        self.last_rssi = 0.0
+        """The RSSI of the last received packet. Stored when the packet was received.
+           The instantaneous RSSI value may not be accurate once the
+           operating mode has been changed.
+        """
+        self.last_snr = 0.0
+        """The SNR of the last received packet. Stored when the packet was received.
+           The instantaneous SNR value may not be accurate once the
+           operating mode has been changed.
+        """
+        # initialize timeouts and delays delays
+        self.ack_wait = 0.5
+        """The delay time before attempting a retry after not receiving an ACK"""
+        self.receive_timeout = 0.5
+        """The amount of time to poll for a received packet.
+           If no packet is received, the returned packet will be None
+        """
+        self.xmit_timeout = 2.0
+        """The amount of time to wait for the HW to transmit the packet.
+           This is mainly used to prevent a hang due to a HW issue
+        """
+        self.ack_retries = 5
+        """The number of ACK retries before reporting a failure."""
+        self.ack_delay = None
+        """The delay time before attemting to send an ACK.
+           If ACKs are being missed try setting this to .1 or .2.
+        """
+        # initialize sequence number counter for reliabe datagram mode
+        self.sequence_number = 0
+        # create seen Ids list
+        self.seen_ids = bytearray(256)
+        # initialize packet header
+        # node address - default is broadcast
+        self.node = _RH_BROADCAST_ADDRESS
+        """The default address of this Node. (0-255).
+           If not 255 (0xff) then only packets address to this node will be accepted.
+           First byte of the RadioHead header.
+        """
+        # destination address - default is broadcast
+        self.destination = _RH_BROADCAST_ADDRESS
+        """The default destination address for packet transmissions. (0-255).
+           If 255 (0xff) then any receiving node should accept the packet.
+           Second byte of the RadioHead header.
+        """
+        # ID - contains seq count for reliable datagram mode
+        self.identifier = 0
+        """Automatically set to the sequence number when send_with_ack() used.
+           Third byte of the RadioHead header.
+        """
+        # flags - identifies ack/reetry packet for reliable datagram mode
+        self.flags = 0
+        """Upper 4 bits reserved for use by Reliable Datagram Mode.
+           Lower 4 bits may be used to pass information.
+           Fourth byte of the RadioHead header.
+        """
+        self.crc_error_count = 0
 
     # pylint: disable=no-member
     # Reconsider pylint: disable when this can be tested
@@ -455,9 +404,9 @@ class RFM9x:
     def reset(self):
         """Perform a reset of the chip."""
         # See section 7.2.2 of the datasheet for reset description.
-        self._reset.switch_to_output(value=False)
+        self._reset.value = False  # Set Reset Low
         time.sleep(0.0001)  # 100 us
-        self._reset.switch_to_input(pull=digitalio.Pull.UP)
+        self._reset.value = True  # set Reset High
         time.sleep(0.005)  # 5 ms
 
     def idle(self):
@@ -566,7 +515,22 @@ class RFM9x:
         """The received strength indicator (in dBm) of the last received message."""
         # Read RSSI register and convert to value using formula in datasheet.
         # Remember in LoRa mode the payload register changes function to RSSI!
-        return self._read_u8(_RH_RF95_REG_1A_PKT_RSSI_VALUE) - 137
+        raw_rssi = self._read_u8(_RH_RF95_REG_1A_PKT_RSSI_VALUE)
+        if self.low_frequency_mode:
+            raw_rssi -= 157
+        else:
+            raw_rssi -= 164
+        return raw_rssi
+
+    @property
+    def snr(self):
+        """The SNR (in dB) of the last received message."""
+        # Read SNR 0x19 register and convert to value using formula in datasheet.
+        # SNR(dB) = PacketSnr [twos complement] / 4
+        snr_byte = self._read_u8(_RH_RF95_REG_19_PKT_SNR_VALUE)
+        if snr_byte > 127:
+            snr_byte = (256 - snr_byte) * -1
+        return snr_byte / 4
 
     @property
     def signal_bandwidth(self):
@@ -593,6 +557,28 @@ class RFM9x:
             _RH_RF95_REG_1D_MODEM_CONFIG1,
             (self._read_u8(_RH_RF95_REG_1D_MODEM_CONFIG1) & 0x0F) | (bw_id << 4),
         )
+        if val >= 500000:
+            # see Semtech SX1276 errata note 2.3
+            self.auto_ifon = True
+            # see Semtech SX1276 errata note 2.1
+            if self.low_frequency_mode:
+                self._write_u8(0x36, 0x02)
+                self._write_u8(0x3A, 0x7F)
+            else:
+                self._write_u8(0x36, 0x02)
+                self._write_u8(0x3A, 0x64)
+        else:
+            # see Semtech SX1276 errata note 2.3
+            self.auto_ifon = False
+            self._write_u8(0x36, 0x03)
+            if val == 7800:
+                self._write_u8(0x2F, 0x48)
+            elif val >= 62500:
+                # see Semtech SX1276 errata note 2.3
+                self._write_u8(0x2F, 0x40)
+            else:
+                self._write_u8(0x2F, 0x44)
+            self._write_u8(0x30, 0)
 
     @property
     def coding_rate(self):
@@ -627,7 +613,12 @@ class RFM9x:
     def spreading_factor(self, val):
         # Set spreading factor (set to 7 to match RadioHead Sf128).
         val = min(max(val, 6), 12)
-        self._write_u8(_RH_RF95_DETECTION_OPTIMIZE, 0xC5 if val == 6 else 0xC3)
+
+        if val == 6:
+            self.detection_optimize = 0x5
+        else:
+            self.detection_optimize = 0x3
+
         self._write_u8(_RH_RF95_DETECTION_THRESHOLD, 0x0C if val == 6 else 0x0A)
         self._write_u8(
             _RH_RF95_REG_1E_MODEM_CONFIG2,
@@ -658,22 +649,41 @@ class RFM9x:
                 self._read_u8(_RH_RF95_REG_1E_MODEM_CONFIG2) & 0xFB,
             )
 
+    def tx_done(self):
+        """Transmit status"""
+        return (self._read_u8(_RH_RF95_REG_12_IRQ_FLAGS) & 0x8) >> 3
+
+    def rx_done(self):
+        """Receive status"""
+        return (self._read_u8(_RH_RF95_REG_12_IRQ_FLAGS) & 0x40) >> 6
+
+    def crc_error(self):
+        """crc status"""
+        return (self._read_u8(_RH_RF95_REG_12_IRQ_FLAGS) & 0x20) >> 5
+
+    # pylint: disable=too-many-branches
     def send(
         self,
         data,
-        timeout=2.0,
+        *,
         keep_listening=False,
-        tx_header=(_RH_BROADCAST_ADDRESS, _RH_BROADCAST_ADDRESS, 0, 0),
+        destination=None,
+        node=None,
+        identifier=None,
+        flags=None
     ):
         """Send a string of data using the transmitter.
-           You can only send 252 bytes at a time
-           (limited by chip's FIFO size and appended headers).
-           This appends a 4 byte header to be compatible with the RadioHead library.
-           The tx_header defaults to using the Broadcast addresses. It may be overidden
-           by specifying a 4-tuple of bytes containing (To,From,ID,Flags)
-           The timeout is just to prevent a hang (arbitrarily set to 2 seconds)
-           The keep_listening argument should be set to True if you want to start listening
-           automatically after the packet is sent. The default setting is False.
+        You can only send 252 bytes at a time
+        (limited by chip's FIFO size and appended headers).
+        This appends a 4 byte header to be compatible with the RadioHead library.
+        The header defaults to using the initialized attributes:
+        (destination,node,identifier,flags)
+        It may be temporarily overidden via the kwargs - destination,node,identifier,flags.
+        Values passed via kwargs do not alter the attribute settings.
+        The keep_listening argument should be set to True if you want to start listening
+        automatically after the packet is sent. The default setting is False.
+
+        Returns: True if success or False if the send timed out.
         """
         # Disable pylint warning to not use length as a check for zero.
         # This is a puzzling warning as the below code is clearly the most
@@ -681,29 +691,48 @@ class RFM9x:
         # buffer be within an expected range of bounds. Disable this check.
         # pylint: disable=len-as-condition
         assert 0 < len(data) <= 252
-        assert len(tx_header) == 4, "tx header must be 4-tuple (To,From,ID,Flags)"
         # pylint: enable=len-as-condition
         self.idle()  # Stop receiving to clear FIFO and keep it clear.
         # Fill the FIFO with a packet to send.
         self._write_u8(_RH_RF95_REG_0D_FIFO_ADDR_PTR, 0x00)  # FIFO starts at 0.
-        # Write header bytes.
-        self._write_u8(_RH_RF95_REG_00_FIFO, tx_header[0])  # Header: To
-        self._write_u8(_RH_RF95_REG_00_FIFO, tx_header[1])  # Header: From
-        self._write_u8(_RH_RF95_REG_00_FIFO, tx_header[2])  # Header: Id
-        self._write_u8(_RH_RF95_REG_00_FIFO, tx_header[3])  # Header: Flags
+        # Combine header and data to form payload
+        payload = bytearray(4)
+        if destination is None:  # use attribute
+            payload[0] = self.destination
+        else:  # use kwarg
+            payload[0] = destination
+        if node is None:  # use attribute
+            payload[1] = self.node
+        else:  # use kwarg
+            payload[1] = node
+        if identifier is None:  # use attribute
+            payload[2] = self.identifier
+        else:  # use kwarg
+            payload[2] = identifier
+        if flags is None:  # use attribute
+            payload[3] = self.flags
+        else:  # use kwarg
+            payload[3] = flags
+        payload = payload + data
         # Write payload.
-        self._write_from(_RH_RF95_REG_00_FIFO, data)
+        self._write_from(_RH_RF95_REG_00_FIFO, payload)
         # Write payload and header length.
-        self._write_u8(_RH_RF95_REG_22_PAYLOAD_LENGTH, len(data) + 4)
+        self._write_u8(_RH_RF95_REG_22_PAYLOAD_LENGTH, len(payload))
         # Turn on transmit mode to send out the packet.
         self.transmit()
         # Wait for tx done interrupt with explicit polling (not ideal but
         # best that can be done right now without interrupts).
-        start = time.monotonic()
         timed_out = False
-        while not timed_out and not self.tx_done:
-            if (time.monotonic() - start) >= timeout:
-                timed_out = True
+        if HAS_SUPERVISOR:
+            start = supervisor.ticks_ms()
+            while not timed_out and not self.tx_done():
+                if ticks_diff(supervisor.ticks_ms(), start) >= self.xmit_timeout * 1000:
+                    timed_out = True
+        else:
+            start = time.monotonic()
+            while not timed_out and not self.tx_done():
+                if time.monotonic() - start >= self.xmit_timeout:
+                    timed_out = True
         # Listen again if necessary and return the result packet.
         if keep_listening:
             self.listen()
@@ -712,74 +741,146 @@ class RFM9x:
             self.idle()
         # Clear interrupt.
         self._write_u8(_RH_RF95_REG_12_IRQ_FLAGS, 0xFF)
-        if timed_out:
-            raise RuntimeError("Timeout during packet send")
+        return not timed_out
+
+    def send_with_ack(self, data):
+        """Reliable Datagram mode:
+        Send a packet with data and wait for an ACK response.
+        The packet header is automatically generated.
+        If enabled, the packet transmission will be retried on failure
+        """
+        if self.ack_retries:
+            retries_remaining = self.ack_retries
+        else:
+            retries_remaining = 1
+        got_ack = False
+        self.sequence_number = (self.sequence_number + 1) & 0xFF
+        while not got_ack and retries_remaining:
+            self.identifier = self.sequence_number
+            self.send(data, keep_listening=True)
+            # Don't look for ACK from Broadcast message
+            if self.destination == _RH_BROADCAST_ADDRESS:
+                got_ack = True
+            else:
+                # wait for a packet from our destination
+                ack_packet = self.receive(timeout=self.ack_wait, with_header=True)
+                if ack_packet is not None:
+                    if ack_packet[3] & _RH_FLAGS_ACK:
+                        # check the ID
+                        if ack_packet[2] == self.identifier:
+                            got_ack = True
+                            break
+            # pause before next retry -- random delay
+            if not got_ack:
+                # delay by random amount before next try
+                time.sleep(self.ack_wait + self.ack_wait * random.random())
+            retries_remaining = retries_remaining - 1
+            # set retry flag in packet header
+            self.flags |= _RH_FLAGS_RETRY
+        self.flags = 0  # clear flags
+        return got_ack
 
     def receive(
-        self,
-        timeout=0.5,
-        keep_listening=True,
-        with_header=False,
-        rx_filter=_RH_BROADCAST_ADDRESS,
+        self, *, keep_listening=True, with_header=False, with_ack=False, timeout=None
     ):
-        """Wait to receive a packet from the receiver. Will wait for up to timeout_s amount of
-           seconds for a packet to be received and decoded. If a packet is found the payload bytes
-           are returned, otherwise None is returned (which indicates the timeout elapsed with no
-           reception).  If timeout is None it is not used ( for use with interrupts)
-           If keep_listening is True (the default) the chip will immediately enter listening mode
-           after reception of a packet, otherwise it will fall back to idle mode and ignore any
-           future reception.
-           A 4-byte header must be prepended to the data for compatibilty with the
-           RadioHead library.
-           The header consists of a 4 bytes (To,From,ID,Flags). The default setting will accept
-           any  incomming packet and strip the header before returning the packet to the caller.
-           If with_header is True then the 4 byte header will be returned with the packet.
-           The payload then begins at packet[4].
-           rx_fliter may be set to reject any "non-broadcast" packets that do not contain the
-           specfied "To" value in the header.
-           if rx_filter is set to 0xff (_RH_BROADCAST_ADDRESS) or if the  "To" field (packet[[0])
-           is equal to 0xff then the packet will be accepted and returned to the caller.
-           If rx_filter is not 0xff and packet[0] does not match rx_filter then
-           the packet is ignored and None is returned.
+        """Wait to receive a packet from the receiver. If a packet is found the payload bytes
+        are returned, otherwise None is returned (which indicates the timeout elapsed with no
+        reception).
+        If keep_listening is True (the default) the chip will immediately enter listening mode
+        after reception of a packet, otherwise it will fall back to idle mode and ignore any
+        future reception.
+        All packets must have a 4-byte header for compatibilty with the
+        RadioHead library.
+        The header consists of 4 bytes (To,From,ID,Flags). The default setting will  strip
+        the header before returning the packet to the caller.
+        If with_header is True then the 4 byte header will be returned with the packet.
+        The payload then begins at packet[4].
+        If with_ack is True, send an ACK after receipt (Reliable Datagram mode)
         """
         timed_out = False
+        if timeout is None:
+            timeout = self.receive_timeout
         if timeout is not None:
-            # Make sure we are listening for packets.
-            self.listen()
-            # Wait for the rx done interrupt.  This is not ideal and will
+            # Wait for the payload_ready signal.  This is not ideal and will
             # surely miss or overflow the FIFO when packets aren't read fast
             # enough, however it's the best that can be done from Python without
             # interrupt supports.
-            start = time.monotonic()
-            while not timed_out and not self.rx_done:
-                if (time.monotonic() - start) >= timeout:
-                    timed_out = True
+            # Make sure we are listening for packets.
+            self.listen()
+            timed_out = False
+            if HAS_SUPERVISOR:
+                start = supervisor.ticks_ms()
+                while not timed_out and not self.rx_done():
+                    if ticks_diff(supervisor.ticks_ms(), start) >= timeout * 1000:
+                        timed_out = True
+            else:
+                start = time.monotonic()
+                while not timed_out and not self.rx_done():
+                    if time.monotonic() - start >= timeout:
+                        timed_out = True
         # Payload ready is set, a packet is in the FIFO.
         packet = None
+        # save last RSSI reading
+        self.last_rssi = self.rssi
+
+        # save the last SNR reading
+        self.last_snr = self.snr
+
+        # Enter idle mode to stop receiving other packets.
+        self.idle()
         if not timed_out:
-            if self.enable_crc and self.crc_error:
-                warn("CRC error, packet ignored")
+            if self.enable_crc and self.crc_error():
+                self.crc_error_count += 1
             else:
-                # Grab the length of the received packet and check it has at least 5
-                # bytes to indicate the 4 byte header and at least 1 byte of user data.
-                length = self._read_u8(_RH_RF95_REG_13_RX_NB_BYTES)
-                if length < 5:
-                    packet = None
-                else:
-                    # Have a good packet, grab it from the FIFO.
-                    # Reset the fifo read ptr to the beginning of the packet.
+                # Read the data from the FIFO.
+                # Read the length of the FIFO.
+                fifo_length = self._read_u8(_RH_RF95_REG_13_RX_NB_BYTES)
+                # Handle if the received packet is too small to include the 4 byte
+                # RadioHead header and at least one byte of data --reject this packet and ignore it.
+                if fifo_length > 0:  # read and clear the FIFO if anything in it
                     current_addr = self._read_u8(_RH_RF95_REG_10_FIFO_RX_CURRENT_ADDR)
                     self._write_u8(_RH_RF95_REG_0D_FIFO_ADDR_PTR, current_addr)
-                    packet = bytearray(length)
+                    packet = bytearray(fifo_length)
                     # Read the packet.
                     self._read_into(_RH_RF95_REG_00_FIFO, packet)
+                # Clear interrupt.
+                self._write_u8(_RH_RF95_REG_12_IRQ_FLAGS, 0xFF)
+                if fifo_length < 5:
+                    packet = None
+                else:
                     if (
-                        rx_filter != _RH_BROADCAST_ADDRESS
+                        self.node != _RH_BROADCAST_ADDRESS
                         and packet[0] != _RH_BROADCAST_ADDRESS
-                        and packet[0] != rx_filter
+                        and packet[0] != self.node
                     ):
                         packet = None
-                    elif not with_header:  # skip the header if not wanted
+                    # send ACK unless this was an ACK or a broadcast
+                    elif (
+                        with_ack
+                        and ((packet[3] & _RH_FLAGS_ACK) == 0)
+                        and (packet[0] != _RH_BROADCAST_ADDRESS)
+                    ):
+                        # delay before sending Ack to give receiver a chance to get ready
+                        if self.ack_delay is not None:
+                            time.sleep(self.ack_delay)
+                        # send ACK packet to sender (data is b'!')
+                        self.send(
+                            b"!",
+                            destination=packet[1],
+                            node=packet[0],
+                            identifier=packet[2],
+                            flags=(packet[3] | _RH_FLAGS_ACK),
+                        )
+                        # reject Retries if we have seen this idetifier from this source before
+                        if (self.seen_ids[packet[1]] == packet[2]) and (
+                            packet[3] & _RH_FLAGS_RETRY
+                        ):
+                            packet = None
+                        else:  # save the packet identifier for this source
+                            self.seen_ids[packet[1]] = packet[2]
+                    if (
+                        not with_header and packet is not None
+                    ):  # skip the header if not wanted
                         packet = packet[4:]
         # Listen again if necessary and return the result packet.
         if keep_listening:
